#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import logging
import re
from typing import Any, Dict, Optional, List

from .base_service import BaseService
from ..types import Device, PropertyIDs, DeviceTypes
from ..utils import create_pid_pair

_LOGGER = logging.getLogger(__name__)


class Bulb(Device):
    _brightness: int = 0
    _color_temp: int = 1800
    _color: Optional[str]
    enr: str

    on: bool = False

    def __init__(self, dictionary: Dict[Any, Any]):
        super().__init__(dictionary)

        self.ip = self.device_params["ip"]

        if self.type is DeviceTypes.MESH_LIGHT:
            self._color = "000000"

    @property
    def brightness(self) -> int:
        return self._brightness

    @brightness.setter
    def brightness(self, value: int) -> None:
        assert value <= 100
        assert value >= 0
        self._brightness = value

    @property
    def color_temp(self) -> int:
        return self._color_temp

    @color_temp.setter
    def color_temp(self, value: int) -> None:
        self._color_temp = value

    @property
    def color(self) -> Optional[str]:
        return self._color

    @color.setter
    def color(self, value) -> None:
        assert re.match(r"^([A-Fa-f\d]{6}|[A-Fa-f\d]{3})$", value) is not None
        self._color = value


class BulbService(BaseService):
    async def update(self, bulb: Bulb) -> Bulb:
        # Get updated device_params
        async with BaseService._update_lock:
            bulb.device_params = await self.get_updated_params(bulb.mac)

        device_info = await self._get_property_list(bulb)
        for property_id, value in device_info:
            if property_id == PropertyIDs.BRIGHTNESS:
                bulb.brightness = int(value)
            elif property_id == PropertyIDs.COLOR_TEMP:
                try:
                    bulb.color_temp = int(value)
                except ValueError:
                    bulb.color_temp = 2700
            elif property_id == PropertyIDs.ON:
                bulb.on = value == "1"
            elif property_id == PropertyIDs.AVAILABLE:
                bulb.available = value == "1"
            elif bulb.type is DeviceTypes.MESH_LIGHT and property_id == PropertyIDs.COLOR:
                bulb.color = value
        return bulb

    async def get_bulbs(self) -> List[Bulb]:
        if self._devices is None:
            self._devices = await self.get_object_list()

        bulbs = [device for device in self._devices if device.type is DeviceTypes.LIGHT or
                 device.type is DeviceTypes.MESH_LIGHT]

        return [Bulb(bulb.raw_dict) for bulb in bulbs]

    async def turn_on(self, bulb: Bulb, options=None):
        plist = [
            create_pid_pair(PropertyIDs.ON, "1")
        ]
        if options is not None:
            plist.extend(options)

        if bulb.type in [
            DeviceTypes.LIGHT
        ]:
            await self._set_property_list(bulb, plist)
        elif bulb.type in [
            DeviceTypes.MESH_LIGHT
        ]:
            await self._local_bulb_command(bulb, plist)

    async def turn_off(self, bulb: Bulb):
        plist = [
            create_pid_pair(PropertyIDs.ON, "0")
        ]

        if bulb.type in [
            DeviceTypes.LIGHT
        ]:
            await self._set_property_list(bulb, plist)
        elif bulb.type in [
            DeviceTypes.MESH_LIGHT
        ]:
            await self._local_bulb_command(bulb, plist)

    async def set_color_temp(self, bulb: Bulb, color_temp: int):
        plist = [
            create_pid_pair(PropertyIDs.COLOR_TEMP, str(color_temp))
        ]

        if bulb.type in [
            DeviceTypes.LIGHT
        ]:
            await self._set_property_list(bulb, plist)
        elif bulb.type in [
            DeviceTypes.MESH_LIGHT
        ]:
            await self._local_bulb_command(bulb, plist)

    async def set_color(self, bulb: Bulb, color: str):
        plist = [
            create_pid_pair(PropertyIDs.COLOR, str(color))
        ]

        if bulb.type in [
            DeviceTypes.MESH_LIGHT
        ]:
            await self._local_bulb_command(bulb, plist)

    async def set_brightness(self, bulb: Device, brightness: int):
        plist = [
            create_pid_pair(PropertyIDs.BRIGHTNESS, str(brightness))
        ]

        if bulb.type in [
            DeviceTypes.LIGHT
        ]:
            await self._set_property_list(bulb, plist)
        if bulb.type in [
            DeviceTypes.MESH_LIGHT
        ]:
            await self._local_bulb_command(bulb, plist)



from asyncio import sleep
from dataclasses import dataclass, field
from heapq import heappush, heappop
from typing import Any
from math import ceil
from ..types import Device

INTERVAL = 300
MAX_SLOTS = 225

@dataclass(order=True)
class DeviceUpdater(object):
    device: Device=field(compare=False)  # The device that will be updated
    service: Any=field(compare=False)
    update_in: int # A countdown to zero that will tell the priority queue that it is time to update this device
    updates_per_interval: int=field(compare=False) # The number of updates that should happen every 5 minutes

    def __init__(self, service, device: Device, update_interval: int):
        """
        This function initializes a DeviceUpdater object
        :param service: The WyzeApy service connected to a device
        :param device: A WyzeApy device that needs to be in the update que
        :param update_interval: How many seconds should be targeted between updates. **Note this value may shift based on 1 call per sec and load.
        """
        self.service = service
        self.device = device
        self.update_in = 0  # Always initialize at 0 so that we get the first update ASAP. The items will shift based on priority after this.
        self.updates_per_interval = ceil(INTERVAL / update_interval)

    async def update(self):
        # We only want to update if the update_in counter is zero
        if self.update_in <= 0:
            # Once it reaches zero and we update the device we want to reset the update_in counter
            self.update_in = ceil(INTERVAL / self.updates_per_interval)
            # Get the updated info for the device from Wyze's API
            self.device = await self.service.update(self.device)
            # Callback to provide the updated info to the subscriber
            self.device.callback_function(self.device)
        else:
            # Don't update and instead just reduce the counter by 1
            self.tick_tock()

    def tick_tock(self):
        # Every time we update a device we want to reduce the update_in counter so that it will get closer to updating
        if self.update_in > 0:
            self.update_in -= 1

    def delay(self):
        # This should be called to reduce the number of updates per interval so that new devices can be added into the queue fairly
        if self.updates_per_interval > 1:
            self.updates_per_interval -= 1

class UpdateManager:
    # Holds all the logic for when to update the devices
    updaters = []
    removed_updaters = []

    def check_if_removed(self, updater: DeviceUpdater):
        for item in self.removed_updaters:
            if updater is item:
                return True
        return False

    # This function should be called once every second
    async def update_next(self):
        while True:
            # First we get the next updater off the queue
            updater = heappop(self.updaters)
            # if the updater has been removed, pop the next and clear it from the removed updaters
            while self.check_if_removed(updater):
                updater = heappop(self.updaters)
                self.removed_updaters.remove(updater)
            # We then reduce the counter for all the other updaters
            self.tick_tock()
            # Then we update the target device
            await updater.update() # It will only update if it is time for it to update. Otherwise it just reduces its update_in counter.
            # Then we put it back at the end of the queue. Or the front again if it wasn't ready to update
            heappush(self.updaters, updater)
            await sleep(1)


    def filled_slots(self):
        # This just returns the number of available slots
        current_slots = 0
        for a_updater in self.updaters:
            current_slots += a_updater.updates_per_interval

        return current_slots

    def decrease_updates_per_interval(self):
        # This will add a delay for all devices so we can squeeze more in there
        for a_updater in self.updaters:
            a_updater.delay()

    def tick_tock(self):
        # This will reduce the update_in counter for all devices
        for a_updater in self.updaters:
            a_updater.tick_tock()

    def add_updater(self, updater: DeviceUpdater):
        if len(self.updaters) >= MAX_SLOTS:
            raise Exception("No more devices can be updated within the rate limit")

        # When we add a new updater it has to fit within the max slots or we will not add it
        while (self.filled_slots() + updater.updates_per_interval) > MAX_SLOTS:
            # If we are overflowing the available slots we will reduce the frequency of updates evenly for all devices until we can fit in one more.
            self.decrease_updates_per_interval()
            updater.delay()

        # Once it fits we will add the new updater to the queue
        heappush(self.updaters, updater)

    def del_updater(self, updater: DeviceUpdater):
        self.removed_updaters.append(updater)



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import logging
from enum import Enum
from typing import Any, Dict, List

from .base_service import BaseService
from ..types import Device, ThermostatProps, DeviceTypes

_LOGGER = logging.getLogger(__name__)


class HVACMode(Enum):
    AUTO = "auto"
    HEAT = "heat"
    COOL = "cool"
    OFF = "off"


class FanMode(Enum):  # auto, on
    AUTO = "auto"
    ON = "on"


class TemperatureUnit(Enum):
    FAHRENHEIT = "F"
    CELSIUS = "C"


class Preset(Enum):
    HOME = "home"
    AWAY = "away"
    SLEEP = "sleep"


class HVACState(Enum):
    COOLING = 'cooling'
    HEATING = 'heating'
    IDLE = 'idle'


class Thermostat(Device):
    def __init__(self, dictionary: Dict[Any, Any]):
        super().__init__(dictionary)

        self.temp_unit: TemperatureUnit = TemperatureUnit.FAHRENHEIT
        self.cool_set_point: int = 74
        self.heat_set_point: int = 64
        self.fan_mode: FanMode = FanMode.AUTO
        self.hvac_mode: HVACMode = HVACMode.AUTO
        self.temperature: float = 71
        self.available: bool = True
        self.humidity: int = 50
        self.hvac_state: HVACState = HVACState.IDLE


class ThermostatService(BaseService):
    async def update(self, thermostat: Thermostat) -> Thermostat:
        properties = (await self._thermostat_get_iot_prop(thermostat))['data']['props']

        device_props = []
        for property in properties:
            try:
                prop = ThermostatProps(property)
                device_props.append((prop, properties[property]))
            except ValueError as e:
                _LOGGER.debug(f"{e} with value {properties[property]}")

        thermostat_props = device_props
        for prop, value in thermostat_props:
            if prop == ThermostatProps.TEMP_UNIT:
                thermostat.temp_unit = TemperatureUnit(value)
            elif prop == ThermostatProps.COOL_SP:
                thermostat.cool_set_point = int(value)
            elif prop == ThermostatProps.HEAT_SP:
                thermostat.heat_set_point = int(value)
            elif prop == ThermostatProps.FAN_MODE:
                thermostat.fan_mode = FanMode(value)
            elif prop == ThermostatProps.MODE_SYS:
                thermostat.hvac_mode = HVACMode(value)
            elif prop == ThermostatProps.TEMPERATURE:
                thermostat.temperature = float(value)
            elif prop == ThermostatProps.IOT_STATE:
                thermostat.available = value == 'connected'
            elif prop == ThermostatProps.HUMIDITY:
                thermostat.humidity = int(value)
            elif prop == ThermostatProps.WORKING_STATE:
                thermostat.hvac_state = HVACState(value)

        return thermostat

    async def get_thermostats(self) -> List[Thermostat]:
        if self._devices is None:
            self._devices = await self.get_object_list()

        thermostats = [device for device in self._devices if device.type is DeviceTypes.THERMOSTAT]

        return [Thermostat(thermostat.raw_dict) for thermostat in thermostats]

    async def set_cool_point(self, thermostat: Device, temp: int):
        await self._thermostat_set_iot_prop(thermostat, ThermostatProps.COOL_SP, temp)

    async def set_heat_point(self, thermostat: Device, temp: int):
        await self._thermostat_set_iot_prop(thermostat, ThermostatProps.HEAT_SP, temp)

    async def set_hvac_mode(self, thermostat: Device, hvac_mode: HVACMode):
        await self._thermostat_set_iot_prop(thermostat, ThermostatProps.MODE_SYS, hvac_mode.value)

    async def set_fan_mode(self, thermostat: Device, fan_mode: FanMode):
        await self._thermostat_set_iot_prop(thermostat, ThermostatProps.FAN_MODE, fan_mode.value)

    async def set_preset(self, thermostat: Thermostat, preset: Preset):
        await self._thermostat_set_iot_prop(thermostat, ThermostatProps.CONFIG_SCENARIO, preset.value)




















#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import asyncio
import logging
from threading import Thread
from typing import List, Callable, Tuple, Optional

from aiohttp import ClientOSError, ContentTypeError

from ..exceptions import UnknownApiError
from .base_service import BaseService
from ..types import Device, PropertyIDs, DeviceTypes

_LOGGER = logging.getLogger(__name__)


class Sensor(Device):
    detected: bool = False


class SensorService(BaseService):
    _updater_thread: Optional[Thread] = None
    _subscribers: List[Tuple[Sensor, Callable[[Sensor], None]]] = []

    async def update(self, sensor: Sensor) -> Sensor:
        # Get updated device_params
        async with BaseService._update_lock:
            sensor.device_params = await self.get_updated_params(sensor.mac)
        properties = await self._get_device_info(sensor)

        for property in properties['data']['property_list']:
            pid = property['pid']
            value = property['value']

            try:
                if PropertyIDs(pid) == PropertyIDs.CONTACT_STATE:
                    sensor.detected = value == "1"
                if PropertyIDs(pid) == PropertyIDs.MOTION_STATE:
                    sensor.detected = value == "1"
            except ValueError:
                pass

        return sensor

    async def register_for_updates(self, sensor: Sensor, callback: Callable[[Sensor], None]):
        _LOGGER.debug(f"Registering sensor: {sensor.nickname} for updates")
        loop = asyncio.get_event_loop()
        if not self._updater_thread:
            self._updater_thread = Thread(target=self.update_worker, args=[loop, ], daemon=True)
            self._updater_thread.start()

        self._subscribers.append((sensor, callback))

    async def deregister_for_updates(self, sensor: Sensor):
        self._subscribers = [(sense, callback) for sense, callback in self._subscribers if sense.mac != sensor.mac]

    def update_worker(self, loop):
        while True:
            for sensor, callback in self._subscribers:
                _LOGGER.debug(f"Providing update for {sensor.nickname}")
                try:
                    callback(asyncio.run_coroutine_threadsafe(self.update(sensor), loop).result())
                except UnknownApiError as e:
                    _LOGGER.warning(f"The update method detected an UnknownApiError: {e}")
                except ClientOSError as e:
                    _LOGGER.error(f"A network error was detected: {e}")
                except ContentTypeError as e:
                    _LOGGER.error(f"Server returned unexpected ContentType: {e}")

    async def get_sensors(self) -> List[Sensor]:
        if self._devices is None:
            self._devices = await self.get_object_list()

        sensors = [Sensor(device.raw_dict) for device in self._devices if
                   device.type is DeviceTypes.MOTION_SENSOR or
                   device.type is DeviceTypes.CONTACT_SENSOR]
        return [Sensor(sensor.raw_dict) for sensor in sensors]



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import asyncio
import logging
import time
from threading import Thread
from typing import Any, List, Optional, Dict, Callable, Tuple

from aiohttp import ClientOSError, ContentTypeError

from ..exceptions import UnknownApiError
from .base_service import BaseService
from .update_manager import DeviceUpdater
from ..types import Device, DeviceTypes, Event, PropertyIDs
from ..utils import return_event_for_device, create_pid_pair

_LOGGER = logging.getLogger(__name__)


class Camera(Device):
    def __init__(self, dictionary: Dict[Any, Any]):
        super().__init__(dictionary)

        self.last_event: Optional[Event] = None
        self.last_event_ts: int = int(time.time() * 1000)
        self.on: bool = True


class CameraService(BaseService):
    _updater_thread: Optional[Thread] = None
    _subscribers: List[Tuple[Camera, Callable[[Camera], None]]] = []

    async def update(self, camera: Camera):
        # Get updated device_params
        async with BaseService._update_lock:
            camera.device_params = await self.get_updated_params(camera.mac)

        # Get camera events
        response = await self._get_event_list(10)
        raw_events = response['data']['event_list']
        latest_events = [Event(raw_event) for raw_event in raw_events]

        if (event := return_event_for_device(camera, latest_events)) is not None:
            camera.last_event = event
            camera.last_event_ts = event.event_ts

        # Update camera state
        state_response: List[Tuple[PropertyIDs, Any]] = await self._get_property_list(camera)
        for property, value in state_response:
            if property is PropertyIDs.AVAILABLE:
                camera.available = value == "1"
            if property is PropertyIDs.ON:
                camera.on = value == "1"

        return camera

    async def register_for_updates(self, camera: Camera, callback: Callable[[Camera], None]):
        loop = asyncio.get_event_loop()
        if not self._updater_thread:
            self._updater_thread = Thread(target=self.update_worker, args=[loop, ], daemon=True)
            self._updater_thread.start()

        self._subscribers.append((camera, callback))

    async def deregister_for_updates(self, camera: Camera):
        self._subscribers = [(cam, callback) for cam, callback in self._subscribers if cam.mac != camera.mac]

    def update_worker(self, loop):
        while True:
            if len(self._subscribers) < 1:
                time.sleep(0.1)
            else:
                for camera, callback in self._subscribers:
                    try:
                        callback(asyncio.run_coroutine_threadsafe(self.update(camera), loop).result())
                    except UnknownApiError as e:
                        _LOGGER.warning(f"The update method detected an UnknownApiError: {e}")
                    except ClientOSError as e:
                        _LOGGER.error(f"A network error was detected: {e}")
                    except ContentTypeError as e:
                        _LOGGER.error(f"Server returned unexpected ContentType: {e}")

    async def get_cameras(self) -> List[Camera]:
        if self._devices is None:
            self._devices = await self.get_object_list()

        cameras = [device for device in self._devices if device.type is DeviceTypes.CAMERA]

        return [Camera(camera.raw_dict) for camera in cameras]

    async def turn_on(self, camera: Camera):
        await self._run_action(camera, "power_on")

    async def turn_off(self, camera: Camera):
        await self._run_action(camera, "power_off")

    async def turn_on_notifications(self, camera: Camera):
        plist = [
            create_pid_pair(PropertyIDs.NOTIFICATION, "1")
        ]

        await self._set_property_list(camera, plist)

    async def turn_off_notifications(self, camera: Camera):
        plist = [
            create_pid_pair(PropertyIDs.NOTIFICATION, "0")
        ]

        await self._set_property_list(camera, plist)

    def register_updater(self, camera: Camera, interval):
        dev_updater = DeviceUpdater(self, camera, interval)
        BaseService._update_manager.add_updater(dev_updater)



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import logging
from inspect import iscoroutinefunction
from typing import List, Optional, Set, Callable

from .const import PHONE_SYSTEM_TYPE, APP_VERSION, SC, APP_VER, SV, PHONE_ID, APP_NAME, OLIVE_APP_ID, APP_INFO
from .crypto import olive_create_signature
from .exceptions import TwoFactorAuthenticationEnabled
from .payload_factory import olive_create_user_info_payload
from .services.base_service import BaseService
from .services.bulb_service import BulbService
from .services.camera_service import CameraService
from .services.hms_service import HMSService
from .services.lock_service import LockService
from .services.sensor_service import SensorService
from .services.switch_service import SwitchService
from .services.thermostat_service import ThermostatService
from .utils import check_for_errors_standard
from .wyze_auth_lib import WyzeAuthLib, Token

_LOGGER = logging.getLogger(__name__)


class Wyzeapy:
    """A module to assist developers in interacting with the Wyze service"""
    # _client: Client
    _auth_lib: WyzeAuthLib

    def __init__(self):
        self._bulb_service = None
        self._switch_service = None
        self._camera_service = None
        self._thermostat_service = None
        self._hms_service = None
        self._lock_service = None
        self._sensor_service = None
        self._email = None
        self._password = None
        self._service: Optional[BaseService] = None
        self._token_callbacks: List[Callable] = []

    @classmethod
    async def create(cls):
        """
        Creates the Wyzeapy class in an async way. Although this is not currently utilized

        :return: An instance of the Wyzeapy class
        """
        self = cls()
        return self

    async def login(self, email, password, token: Token = None):
        """
        Logs the user in and retrieves the users token

        :param email: Users email
        :param password: Users password
        :param token: Users existing token from a previous session

        :raises:
            TwoFactorAuthenticationEnabled: indicates that the account has 2fa enabled
        """

        self._email = email
        self._password = password
        try:
            if token:
                # User token supplied, lets go ahead and use it and refresh the access token if needed.
                self._auth_lib = await WyzeAuthLib.create(
                    email, password, token, token_callback=self.execute_token_callbacks)
                await self._auth_lib.refresh_if_should()
                self._service = BaseService(self._auth_lib)
            else:
                self._auth_lib = await WyzeAuthLib.create(email, password, token_callback=self.execute_token_callbacks)
                await self._auth_lib.get_token_with_username_password(email, password)
                self._service = BaseService(self._auth_lib)
        except TwoFactorAuthenticationEnabled as error:
            raise error

    async def login_with_2fa(self, verification_code) -> Token:
        """
        Logs the user in and retrieves the users token

        :param verification_code: Users 2fa verification code

        """

        _LOGGER.debug(f"Verification Code: {verification_code}")

        await self._auth_lib.get_token_with_2fa(verification_code)
        self._service = BaseService(self._auth_lib)
        return self._auth_lib.token

    async def execute_token_callbacks(self, token: Token):
        """
        Sends the token to the registered callback functions.

        :param token: Users token object

        """
        for callback in self._token_callbacks:
            if iscoroutinefunction(callback):
                await callback(token)
            else:
                callback(token)

    def register_for_token_callback(self, callback_function):
        """
        Register a callback to be called whenever the user's token is modified

        :param callback_function: A callback function which expects a token object

        """
        self._token_callbacks.append(callback_function)

    def unregister_for_token_callback(self, callback_function):
        """
        Register a callback to be called whenever the user's token is modified

        :param callback_function: A callback function which expects a token object

        """
        self._token_callbacks.remove(callback_function)

    @property
    async def unique_device_ids(self) -> Set[str]:
        """
        Returns a list of all device ids known to the server
        :return: A set containing the unique device ids
        """

        devices = await self._service.get_object_list()
        device_ids = set()
        for device in devices:
            device_ids.add(device.mac)

        return device_ids

    @property
    async def notifications_are_on(self) -> bool:
        """
        Reports the status of the notifications

        :return: True if the notifications are enabled
        """

        response_json = await self._service.get_user_profile()
        return response_json['data']['notification']

    async def enable_notifications(self):
        """Enables notifications on the account"""

        await self._service.set_push_info(True)

    async def disable_notifications(self):
        """Disables notifications on the account"""

        await self._service.set_push_info(False)

    @classmethod
    async def valid_login(cls, email: str, password: str) -> bool:
        """
        Checks to see if a username and password return a valid login

        :param email: The users email
        :param password: The users password
        :return: True if the account can connect
        """

        self = cls()
        await self.login(email, password)
        return not self._auth_lib.should_refresh

    @property
    async def bulb_service(self) -> BulbService:
        """Returns an instance of the bulb service"""

        if self._bulb_service is None:
            self._bulb_service = BulbService(self._auth_lib)
        return self._bulb_service

    @property
    async def switch_service(self) -> SwitchService:
        """Returns an instance of the switch service"""

        if self._switch_service is None:
            self._switch_service = SwitchService(self._auth_lib)
        return self._switch_service

    @property
    async def camera_service(self) -> CameraService:
        """Returns an instance of the camera service"""

        if self._camera_service is None:
            self._camera_service = CameraService(self._auth_lib)
        return self._camera_service

    @property
    async def thermostat_service(self) -> ThermostatService:
        """Returns an instance of the thermostat service"""

        if self._thermostat_service is None:
            self._thermostat_service = ThermostatService(self._auth_lib)
        return self._thermostat_service

    @property
    async def hms_service(self) -> HMSService:
        """Returns an instance of the hms service"""

        if self._hms_service is None:
            self._hms_service = await HMSService.create(self._auth_lib)
        return self._hms_service

    @property
    async def lock_service(self) -> LockService:
        """Returns an instance of the lock service"""

        if self._lock_service is None:
            self._lock_service = LockService(self._auth_lib)
        return self._lock_service

    @property
    async def sensor_service(self) -> SensorService:
        """Returns an instance of the sensor service"""

        if self._sensor_service is None:
            self._sensor_service = SensorService(self._auth_lib)
        return self._sensor_service



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import asyncio
import logging
import time
from typing import Dict, Any, Optional

from aiohttp import TCPConnector, ClientSession, ContentTypeError

from .const import API_KEY, PHONE_ID, APP_NAME, APP_VERSION, SC, SV, PHONE_SYSTEM_TYPE, APP_VER, APP_INFO
from .exceptions import UnknownApiError, TwoFactorAuthenticationEnabled
from .utils import create_password, check_for_errors_standard

_LOGGER = logging.getLogger(__name__)


class Token:
    # Token is good for 216,000 seconds (60 hours) but 48 hours seems like a reasonable refresh interval
    REFRESH_INTERVAL = 172800

    def __init__(self, access_token, refresh_token, refresh_time: float = None):
        self._access_token: str = access_token
        self._refresh_token: str = refresh_token
        if refresh_time:
            self._refresh_time: float = refresh_time
        else:
            self._refresh_time: float = time.time() + Token.REFRESH_INTERVAL

    @property
    def access_token(self):
        return self._access_token

    @access_token.setter
    def access_token(self, access_token):
        self._access_token = access_token
        self._refresh_time = time.time() + Token.REFRESH_INTERVAL

    @property
    def refresh_token(self):
        return self._refresh_token

    @refresh_token.setter
    def refresh_token(self, refresh_token):
        self._refresh_token = refresh_token

    @property
    def refresh_time(self):
        return self._refresh_time


class WyzeAuthLib:
    token: Optional[Token] = None
    SANITIZE_FIELDS = [
        "email",
        "password",
        "access_token",
        "accessToken",
        "refresh_token",
        "lat",
        "lon",
        "address",
    ]
    SANITIZE_STRING = "**Sanitized**"

    def __init__(self, username=None, password=None, token: Token = None, token_callback=None):
        self._username = username
        self._password = password
        self.token = token
        self.session_id = ""
        self.verification_id = ""
        self.two_factor_type = None
        self.refresh_lock = asyncio.Lock()
        self.token_callback = token_callback

    @classmethod
    async def create(cls, username=None, password=None, token: Token = None, token_callback=None):
        self = cls(username=username, password=password, token=token, token_callback=token_callback)

        if self._username is None and self._password is None and self.token is None:
            raise AttributeError("Must provide a username, password or token")
        elif self.token is None and self._username is not None and self._password is not None:
            assert self._username != ""
            assert self._password != ""

        return self

    async def get_token_with_username_password(self, username, password) -> Token:
        self._username = username
        self._password = create_password(password)
        login_payload = {
            "email": self._username,
            "password": self._password
        }

        headers = {
            'Phone-Id': PHONE_ID,
            'User-Agent': APP_INFO,
            'X-API-Key': API_KEY,
        }

        response_json = await self.post("https://auth-prod.api.wyze.com/user/login", headers=headers,
                                        json=login_payload)

        if response_json.get('errorCode') is not None:
            _LOGGER.error(f"Unable to login with response from Wyze: {response_json}")
            raise UnknownApiError(response_json)

        if response_json.get('mfa_options') is not None:
            # Store the TOTP verification setting in the token and raise exception
            if "TotpVerificationCode" in response_json.get("mfa_options"):
                self.two_factor_type = "TOTP"
                # Store the verification_id from the response, it's needed for the 2fa payload.
                self.verification_id = response_json["mfa_details"]["totp_apps"][0]["app_id"]
                raise TwoFactorAuthenticationEnabled
                # 2fa using SMS, store sms as 2fa method in token, send the code then raise exception
            if "PrimaryPhone" in response_json.get("mfa_options"):
                self.two_factor_type = "SMS"
                params = {
                    'mfaPhoneType': 'Primary',
                    'sessionId': response_json.get("sms_session_id"),
                    'userId': response_json['user_id'],
                }
                response_json = await self.post('https://auth-prod.api.wyze.com/user/login/sendSmsCode',
                                                headers=headers, data=params)
                # Store the session_id from this response, it's needed for the 2fa payload.
                self.session_id = response_json['session_id']
                raise TwoFactorAuthenticationEnabled

        self.token = Token(response_json['access_token'], response_json['refresh_token'])
        await self.token_callback(self.token)
        return self.token

    async def get_token_with_2fa(self, verification_code) -> Token:
        headers = {
            'Phone-Id': PHONE_ID,
            'User-Agent': APP_INFO,
            'X-API-Key': API_KEY,
        }
        # TOTP Payload
        if self.two_factor_type == "TOTP":
            payload = {
                "email": self._username,
                "password": self._password,
                "mfa_type": "TotpVerificationCode",
                "verification_id": self.verification_id,
                "verification_code": verification_code
            }
        # SMS Payload
        else:
            payload = {
                "email": self._username,
                "password": self._password,
                "mfa_type": "PrimaryPhone",
                "verification_id": self.session_id,
                "verification_code": verification_code
            }

        response_json = await self.post(
            'https://auth-prod.api.wyze.com/user/login',
            headers=headers, json=payload)

        self.token = Token(response_json['access_token'], response_json['refresh_token'])
        await self.token_callback(self.token)
        return self.token

    @property
    def should_refresh(self) -> bool:
        return time.time() >= self.token.refresh_time

    async def refresh_if_should(self):
        if self.should_refresh:
            async with self.refresh_lock:
                if self.should_refresh:
                    _LOGGER.debug("Should refresh. Refreshing...")
                    await self.refresh()

    async def refresh(self) -> None:
        payload = {
            "phone_id": PHONE_ID,
            "app_name": APP_NAME,
            "app_version": APP_VERSION,
            "sc": SC,
            "sv": SV,
            "phone_system_type": PHONE_SYSTEM_TYPE,
            "app_ver": APP_VER,
            "ts": int(time.time()),
            "refresh_token": self.token.refresh_token
        }

        headers = {
            "X-API-Key": API_KEY
        }

        async with ClientSession(connector=TCPConnector(ttl_dns_cache=(30 * 60))) as _session:
            response = await _session.post("https://api.wyzecam.com/app/user/refresh_token", headers=headers,
                                           json=payload)
        response_json = await response.json()
        check_for_errors_standard(response_json)

        self.token.access_token = response_json['data']['access_token']
        self.token.refresh_token = response_json['data']['refresh_token']
        await self.token_callback(self.token)

    def sanitize(self, data):
        if data and type(data) is dict:
            # value is unused, but it prevents us from having to split the tuple to check against SANITIZE_FIELDS
            for key, value in data.items():
                if type(value) is dict:
                    data[key] = self.sanitize(value)
                if key in self.SANITIZE_FIELDS:
                    data[key] = self.SANITIZE_STRING
        return data

    async def post(self, url, json=None, headers=None, data=None) -> Dict[Any, Any]:
        async with ClientSession(connector=TCPConnector(ttl_dns_cache=(30 * 60))) as _session:
            response = await _session.post(url, json=json, headers=headers, data=data)
            # Relocated these below as the sanitization seems to modify the data before it goes to the post.
            _LOGGER.debug("Request:")
            _LOGGER.debug(f"url: {url}")
            _LOGGER.debug(f"json: {self.sanitize(json)}")
            _LOGGER.debug(f"headers: {self.sanitize(headers)}")
            _LOGGER.debug(f"data: {self.sanitize(data)}")
            # Log the response.json() if it exists, if not log the response.
            try:
                response_json = await response.json()
                _LOGGER.debug(f"Response Json: {self.sanitize(response_json)}")
            except ContentTypeError:
                _LOGGER.debug(f"Response: {response}")
            return await response.json()

    async def get(self, url, headers=None, params=None) -> Dict[Any, Any]:
        async with ClientSession(connector=TCPConnector(ttl_dns_cache=(30 * 60))) as _session:
            response = await _session.get(url, params=params, headers=headers)
            # Relocated these below as the sanitization seems to modify the data before it goes to the post.
            _LOGGER.debug("Request:")
            _LOGGER.debug(f"url: {url}")
            _LOGGER.debug(f"headers: {self.sanitize(headers)}")
            _LOGGER.debug(f"params: {self.sanitize(params)}")
            # Log the response.json() if it exists, if not log the response.
            try:
                response_json = await response.json()
                _LOGGER.debug(f"Response Json: {self.sanitize(response_json)}")
            except ContentTypeError:
                _LOGGER.debug(f"Response: {response}")
            return await response.json()

    async def patch(self, url, headers=None, params=None, json=None) -> Dict[Any, Any]:
        async with ClientSession(connector=TCPConnector(ttl_dns_cache=(30 * 60))) as _session:
            response = await _session.patch(url, headers=headers, params=params, json=json)
            # Relocated these below as the sanitization seems to modify the data before it goes to the post.
            _LOGGER.debug("Request:")
            _LOGGER.debug(f"url: {url}")
            _LOGGER.debug(f"json: {self.sanitize(json)}")
            _LOGGER.debug(f"headers: {self.sanitize(headers)}")
            _LOGGER.debug(f"params: {self.sanitize(params)}")
            # Log the response.json() if it exists, if not log the response.
            try:
                response_json = await response.json()
                _LOGGER.debug(f"Response Json: {self.sanitize(response_json)}")
            except ContentTypeError:
                _LOGGER.debug(f"Response: {response}")
            return await response.json()

    async def delete(self, url, headers=None, json=None) -> Dict[Any, Any]:
        async with ClientSession(connector=TCPConnector(ttl_dns_cache=(30 * 60))) as _session:
            response = await _session.delete(url, headers=headers, json=json)
            # Relocated these below as the sanitization seems to modify the data before it goes to the post.
            _LOGGER.debug("Request:")
            _LOGGER.debug(f"url: {url}")
            _LOGGER.debug(f"json: {self.sanitize(json)}")
            _LOGGER.debug(f"headers: {self.sanitize(headers)}")
            # Log the response.json() if it exists, if not log the response.
            try:
                response_json = await response.json()
                _LOGGER.debug(f"Response Json: {self.sanitize(response_json)}")
            except ContentTypeError:
                _LOGGER.debug(f"Response: {response}")
            return await response.json()



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
from typing import Dict, Any


class ActionNotSupported(Exception):
    def __init__(self, device_type: str):
        message = "The action specified is not supported by device type: {}".format(device_type)

        super().__init__(message)


class ParameterError(Exception):
    pass


class AccessTokenError(Exception):
    pass


class LoginError(Exception):
    pass


class UnknownApiError(Exception):
    def __init__(self, response_json: Dict[str, Any]):
        super(UnknownApiError, self).__init__(str(response_json))


class TwoFactorAuthenticationEnabled(Exception):
    pass



#  Copyright (c) 2021. Mulliken, LLC - All Rights Reserved
#  You may use, distribute and modify this code under the terms
#  of the attached license. You should have received a copy of
#  the license with this file. If not, please write to:
#  joshua@mulliken.net to receive a copy
import base64
import hashlib
from typing import Dict, Any, List, Optional

from Crypto.Cipher import AES

from .exceptions import ParameterError, AccessTokenError, UnknownApiError
from .types import ResponseCodes, PropertyIDs, Device, Event

PADDING = bytes.fromhex("05")


def pad(plain_text):
    """
    func to pad cleartext to be multiples of 8-byte blocks.
    If you want to encrypt a text message that is not multiples of 8-byte
    blocks, the text message must be padded with additional bytes to make the
    text message to be multiples of 8-byte blocks.
    """
    raw = plain_text.encode("ascii")

    pad_num = AES.block_size - len(raw) % AES.block_size
    raw += PADDING * pad_num

    return raw


def wyze_encrypt(key, text):
    """
    Reimplementation of the Wyze app's encryption mechanism.

    The decompiled code can be found here 👇
    https://paste.sr.ht/~joshmulliken/e9f67e05c4a774004b226d2ac1f070b6d341cb39
    """
    raw = pad(text)
    key = key.encode("ascii")
    iv = key  # Wyze uses the secret key for the iv as well
    cipher = AES.new(key, AES.MODE_CBC, iv)
    enc = cipher.encrypt(raw)
    b64_enc = base64.b64encode(enc).decode("ascii")
    b64_enc = b64_enc.replace("/", r'\/')
    return b64_enc


def wyze_decrypt(key, enc):
    """
    Reimplementation of the Wyze app's decryption mechanism.

    The decompiled code can be found here 👇
    https://paste.sr.ht/~joshmulliken/e9f67e05c4a774004b226d2ac1f070b6d341cb39
    """
    enc = base64.b64decode(enc)

    key = key.encode('ascii')
    iv = key
    cipher = AES.new(key, AES.MODE_CBC, iv)
    decrypt = cipher.decrypt(enc)

    decrypt_txt = decrypt.decode("ascii")

    return decrypt_txt


def create_password(password: str) -> str:
    hex1 = hashlib.md5(password.encode()).hexdigest()
    hex2 = hashlib.md5(hex1.encode()).hexdigest()
    return hashlib.md5(hex2.encode()).hexdigest()


def check_for_errors_standard(response_json: Dict[str, Any]) -> None:
    if response_json['code'] != ResponseCodes.SUCCESS.value:
        if response_json['code'] == ResponseCodes.PARAMETER_ERROR.value:
            raise ParameterError(response_json)
        elif response_json['code'] == ResponseCodes.ACCESS_TOKEN_ERROR.value:
            raise AccessTokenError
        elif response_json['code'] == ResponseCodes.DEVICE_OFFLINE.value:
            return
        else:
            raise UnknownApiError(response_json)


def check_for_errors_lock(response_json: Dict[str, Any]) -> None:
    if response_json['ErrNo'] != 0:
        if response_json.get('code') == ResponseCodes.PARAMETER_ERROR.value:
            raise ParameterError
        elif response_json.get('code') == ResponseCodes.ACCESS_TOKEN_ERROR.value:
            raise AccessTokenError
        else:
            raise UnknownApiError(response_json)


def check_for_errors_thermostat(response_json: Dict[Any, Any]) -> None:
    if response_json['code'] != 1:
        raise UnknownApiError(response_json)


def check_for_errors_hms(response_json: Dict[Any, Any]) -> None:
    if response_json['message'] is None:
        raise AccessTokenError


def return_event_for_device(device: Device, events: List[Event]) -> Optional[Event]:
    for event in events:
        if event.device_mac == device.mac:
            return event

    return None


def create_pid_pair(pid_enum: PropertyIDs, value: str) -> Dict[str, str]:
    return {"pid": pid_enum.value, "pvalue": value}
